/* QServer.c - NeonServ v5.6
 * Copyright (C) 2011-2012  Philipp Kreil (pk910)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License 
 * along with this program. If not, see <http://www.gnu.org/licenses/>. 
 */

#include "QServer.h"
#include "UserNode.h"
#include "ChanNode.h"
#include "ModeNode.h"
#include "ChanUser.h"
#include "ClientSocket.h"
#include "WHOHandler.h"
#include "ConfigParser.h"
#include "bots.h"
#include "IOHandler.h"

#ifdef WIN32
typedef uint32_t socklen_t;
#endif

#define QSERVER_TIMEOUT 30
#define QSERVER_MAXCLIENTS 100

#define QSERVER_FLAG_AUTHED     0x01
#define QSERVER_FLAG_IN_USE     0x02

struct QServerClient {
    struct IODescriptor *iofd;
    unsigned int flags;
    int references;
    struct QServerClient *next;
};

static struct IODescriptor *server_iofd = NULL;
struct QServerClient *qserver_clients = NULL;
static int qserver_clientcount = 0;

static IOHANDLER_CALLBACK(qserver_callback);

void qserver_init() {
    if(get_int_field("QServer.enabled")) {
        char *host = get_string_field("QServer.host");
        if(!host)
            host = "0.0.0.0";
        int port = get_int_field("QServer.port");
        if(!port)
            port = 7499;
        server_iofd = iohandler_listen(host, port, qserver_callback);
    }
}

void qserver_free() {
    if(!server_iofd)
        return;
    struct QServerClient *client, *next;
    for (client = qserver_clients; client; client = next) {
        next = client->next;
        if(client->iofd)
            iohandler_close(client->iofd);
        free(client);
    }
    qserver_clients = NULL;
    qserver_clientcount = 0;
    iohandler_close(server_iofd);
    server_iofd = NULL;
}

static int qserver_write(struct QServerClient *client, char* msg, int len) {
    if (!client || !client->iofd) return 0;
    if(!len)
        len = strlen(msg);
	iohandler_send(client->iofd, msg, len);
    return 1;
}

static void qserver_put(struct QServerClient *client, const char *text, ...) {
    va_list arg_list;
    char sendBuf[MAXLEN];
    int pos;
    if (!client || !client->iofd) return;
    sendBuf[0] = '\0';
    va_start(arg_list, text);
    pos = vsnprintf(sendBuf, MAXLEN - 2, text, arg_list);
    va_end(arg_list);
    if (pos < 0 || pos > (MAXLEN - 2)) pos = MAXLEN - 2;
    sendBuf[pos] = '\n';
    sendBuf[pos+1] = '\0';
    qserver_write(client, sendBuf, pos+1);
}

static void qserver_update_lastmsg(struct QServerClient *client) {
    struct timeval timeout;
    gettimeofday(&timeout, NULL);
    timeout.tv_sec += QSERVER_TIMEOUT;
    iohandler_set_timeout(client->iofd, &timeout);
}

static void qserver_parse_A(struct QServerClient *client, char **argv, int argc) {
    if(client->flags & QSERVER_FLAG_AUTHED) {
        qserver_put(client, "E :Already Authed");
        return;
    }
    if(!argv) {
        qserver_put(client, "E :Missing Parameter");
        return;
    }
    if(strcmp(argv[0], get_string_field("QServer.pass"))) {
        qserver_put(client, "E :Wrong Password");
        return;
    }
    client->flags |= QSERVER_FLAG_AUTHED;
    qserver_update_lastmsg(client);
    qserver_put(client, "A :Logged in");
}

#define QSERVER_COMMAND_HEADER {\
    if(!(client->flags & QSERVER_FLAG_AUTHED)) {\
        qserver_put(client, "E :Not Authed");\
        return;\
    }\
    qserver_update_lastmsg(client);\
}

static void qserver_parse_U(struct QServerClient *client, char **argv, int argc);
static void qserver_parse_C(struct QServerClient *client, char **argv, int argc);
static void qserver_parse_AC(struct QServerClient *client, char **argv, int argc);
static void qserver_parse_ACU(struct QServerClient *client, char **argv, int argc);
static void qserver_parse_R(struct QServerClient *client, char **argv, int argc);

static void qserver_parse(struct QServerClient *client, char *line) {
    int argc = 0;
    char *argv[MAXNUMPARAMS];
    while(*line) {
        //skip leading spaces
        while (*line == ' ')
            *line++ = 0;
        if (*line == ':') {
           //the rest is a single parameter
           argv[argc++] = line + 1;
           break;
        }
        argv[argc++] = line;
        if (argc >= MAXNUMPARAMS)
            break;
        while (*line != ' ' && *line)
            line++;
    }
    if(!stricmp(argv[0], "A")) //AUTH
        qserver_parse_A(client, argv+1, argc-1);
    else if(!stricmp(argv[0], "U")) //get User
        qserver_parse_U(client, argv+1, argc-1);
    else if(!stricmp(argv[0], "C")) //get Channel
        qserver_parse_C(client, argv+1, argc-1);
    else if(!stricmp(argv[0], "AC")) //get All Channels
        qserver_parse_AC(client, argv+1, argc-1);
    else if(!stricmp(argv[0], "ACU")) //get All ChannelUsers
        qserver_parse_ACU(client, argv+1, argc-1);
    else if(!stricmp(argv[0], "R")) //RAW
        qserver_parse_R(client, argv+1, argc-1);
    else
        qserver_put(client, "E :Unknown Command");
}

static void qserver_accept(int sockfd) {
    struct IODescriptor *client_iofd = iohandler_add(sockfd, IOTYPE_CLIENT, NULL, qserver_callback);
    client_iofd->state = IO_CONNECTED;
    iohandler_update(client_iofd);
    if(qserver_clientcount >= QSERVER_MAXCLIENTS) {
        iohandler_printf(client_iofd, "E :Maximum QServer Connections reached");
        iohandler_close(client_iofd);
        return;
    }
    struct QServerClient *client = malloc(sizeof(*client));
    client->iofd = client_iofd;
    client->references = 0;
    client->next = qserver_clients;
    qserver_clients = client;
    qserver_clientcount++;
}

static void qserver_cleanup() {
    struct QServerClient *client, *next, *prev = NULL;
    for (client = qserver_clients; client; client = next) {
        next = client->next;
        if(client->iofd == NULL && !(client->flags & QSERVER_FLAG_IN_USE)) {
            if(prev) 
                prev->next = client->next;
            else
                qserver_clients = client->next;
            qserver_clientcount--;
            free(client);
        }
    }
}

static IOHANDLER_CALLBACK(qserver_callback) {
    struct QServerClient *client = event->iofd->data;
    switch(event->type) {
    case IOEVENT_TIMEOUT:
        qserver_put(client, "E :Timeout");
        client->iofd = NULL;
        break;
    case IOEVENT_RECV:
        qserver_parse(client, event->data.recv_str);
        break;
    case IOEVENT_CLOSED:
        iohandler_close(client->iofd);
        client->iofd = NULL;
        break;
    case IOEVENT_ACCEPT:
        qserver_accept(event->data.accept_fd);
        break;
    default:
        break;
    }
    qserver_cleanup();
}

/* 
* Command functions
*/

static USERAUTH_CALLBACK(qserver_parse_U_async);
static void qserver_parse_U(struct QServerClient *client, char **argv, int argc) {
    QSERVER_COMMAND_HEADER;
    if(!argv) {
        qserver_put(client, "E :Missing Parameter");
        return;
    }
    struct UserNode *cuser = getUserByNick(argv[0]);
    if(!cuser) {
        cuser = createTempUser(argv[0]);
        if(!cuser) {
            qserver_put(client, "U 0 :Unknown User");
            return;
        }
        cuser->flags |= USERFLAG_ISTMPUSER;
    }
    client->references++;
    client->flags |= QSERVER_FLAG_IN_USE;
    get_userauth(cuser, 0, qserver_parse_U_async, client);
}

static USERAUTH_CALLBACK(qserver_parse_U_async) {
    struct QServerClient *qclient = data;
    qclient->references--;
    if(!qclient->references)
        qclient->flags &= ~QSERVER_FLAG_IN_USE;
    if(!user) {
        qserver_put(qclient, "U 0 :Unknown User");
        return;
    }
    qserver_put(qclient, "U 1 %s %s %s %s :%s", user->nick, user->ident, user->host, ((user->flags & USERFLAG_ISAUTHED) ? user->auth : "0"), user->realname);
}

static void qserver_parse_C(struct QServerClient *client, char **argv, int argc) {
    QSERVER_COMMAND_HEADER;
    if(!argv) {
        qserver_put(client, "E :Missing Parameter");
        return;
    }
    struct ChanNode *chan = getChanByName(argv[0]);
    if(!chan) {
        qserver_put(client, "C 0 :Unknown Channel");
        return;
    }
    char tmpStr[MAXLEN];
    getModeString(chan->modes, tmpStr);
    qserver_put(client, "C 1 %s %d %s :%s", chan->name, chan->usercount, tmpStr, chan->topic);
}

static void qserver_parse_AC(struct QServerClient *client, char **argv, int argc) {
    QSERVER_COMMAND_HEADER;
    struct ChanNode *chan;
    char tmpStr[MAXLEN];
    for(chan = getAllChans(NULL); chan; chan = getAllChans(chan)) {
        getModeString(chan->modes, tmpStr);
        qserver_put(client, "AC %s %d %s :%s", chan->name, chan->usercount, tmpStr, chan->topic);
    }
    qserver_put(client, "ACE"); //AllChannelsEnd
}

static USERLIST_CALLBACK(qserver_parse_ACU_async);
static void qserver_parse_ACU(struct QServerClient *client, char **argv, int argc) {
    QSERVER_COMMAND_HEADER;
    if(!argv) {
        qserver_put(client, "E :Missing Parameter");
        return;
    }
    struct ChanNode *chan = getChanByName(argv[0]);
    if(!chan) {
        qserver_put(client, "ACUE 0 :Unknown Channel");
        return;
    }
    if(argc > 1 && !stricmp(argv[1], "1")) {
        client->references++;
        client->flags |= QSERVER_FLAG_IN_USE;
        get_userlist_if_invisible(chan, 0, qserver_parse_ACU_async, client);
        return;
    }
    char tmpStr[6];
    int tmpStrPos;
    struct ChanUser *chanuser;
    for(chanuser = getChannelUsers(chan, NULL); chanuser; chanuser = getChannelUsers(chan, chanuser)) {
        tmpStrPos = 0;
        if(chanuser->flags & CHANUSERFLAG_OPPED)
            tmpStr[tmpStrPos++] = '@';
        if(chanuser->flags & CHANUSERFLAG_HALFOPPED)
            tmpStr[tmpStrPos++] = '%';
        if(chanuser->flags & CHANUSERFLAG_VOICED)
            tmpStr[tmpStrPos++] = '+';
        if(chanuser->flags & CHANUSERFLAG_INVISIBLE)
            tmpStr[tmpStrPos++] = '<';
        tmpStr[tmpStrPos] = '\0';
        qserver_put(client, "ACU %s %s %s", chanuser->user->nick, ((chanuser->user->flags & USERFLAG_ISAUTHED) ? chanuser->user->auth : "0"), tmpStr);
    }
    qserver_put(client, "ACUE 1");
}

static USERLIST_CALLBACK(qserver_parse_ACU_async) {
    struct QServerClient *qclient = data;
    qclient->references--;
    if(!qclient->references)
        qclient->flags &= ~QSERVER_FLAG_IN_USE;
    char tmpStr[6];
    int tmpStrPos;
    struct ChanUser *chanuser;
    for(chanuser = getChannelUsers(chan, NULL); chanuser; chanuser = getChannelUsers(chan, chanuser)) {
        tmpStrPos = 0;
        if(chanuser->flags & CHANUSERFLAG_OPPED)
            tmpStr[tmpStrPos++] = '@';
        if(chanuser->flags & CHANUSERFLAG_HALFOPPED)
            tmpStr[tmpStrPos++] = '%';
        if(chanuser->flags & CHANUSERFLAG_VOICED)
            tmpStr[tmpStrPos++] = '+';
        if(chanuser->flags & CHANUSERFLAG_INVISIBLE)
            tmpStr[tmpStrPos++] = '<';
        tmpStr[tmpStrPos] = '\0';
        qserver_put(qclient, "ACU %s %s %s", chanuser->user->nick, ((chanuser->user->flags & USERFLAG_ISAUTHED) ? chanuser->user->auth : "0"), tmpStr);
    }
    qserver_put(qclient, "ACUE 1");
}

static void qserver_parse_R(struct QServerClient *client, char **argv, int argc) {
    QSERVER_COMMAND_HEADER;
    if(argc < 3) {
        qserver_put(client, "E :Missing Parameter");
        return;
    }
    struct ClientSocket *bot;
    if(!strcmp(argv[0], "1")) {
        for(bot = getBots(SOCKET_FLAG_READY, NULL); bot; bot = getBots(SOCKET_FLAG_READY, bot)) {
            if(!stricmp(bot->user->nick, argv[1])) break;
        }
    } else {
        struct ClientSocket *low_bot;
        int botid = resolve_botalias(argv[1]);
        if(botid == -1)
            botid = atoi(argv[1]);
        for(bot = getBots(SOCKET_FLAG_READY, NULL); bot; bot = getBots(SOCKET_FLAG_READY, bot)) {
            if(bot->botid == botid) {
                if(bot->flags & SOCKET_FLAG_PREFERRED) break;
                low_bot = bot;
            }
        }
        if(!bot)
            bot = low_bot;
    }
    if(!bot) {
        qserver_put(client, "R 0 :Bot not found");
        return;
    }
    putsock(bot, "%s", argv[2]);
    qserver_put(client, "R 1");
}
