/* main.c - NeonServ v5.6
 * Copyright (C) 2011-2012  Philipp Kreil (pk910)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License 
 * along with this program. If not, see <http://www.gnu.org/licenses/>. 
 */

#include "main.h"
#include "signal.h"
#include "ClientSocket.h"
#include "UserNode.h"
#include "ChanNode.h"
#include "IRCEvents.h"
#include "IRCParser.h"
#include "modcmd.h"
#include "WHOHandler.h"
#include "bots.h"
#include "mysqlConn.h"
#include "HandleInfoHandler.h"
#include "lang.h"
#include "tools.h"
#include "timeq.h"
#include "EventLogger.h"
#include "ModeNode.h"
#include "IRCQueue.h"
#include "DBHelper.h"
#include "ConfigParser.h"
#include "QServer.h"
#include "version.h"
#include "modules.h"
#include "module_commands.h"
#include "ModuleFunctions.h"
#include "IOHandler.h"

time_t start_time;
static int running, hard_restart;
static int statistics_requested_lusers = 0;
int statistics_enabled;
TIMEQ_CALLBACK(main_statistics);
TIMEQ_CALLBACK(main_checkauths);
static int daemonized = 0;
static int print_loglevel = 0;
static FILE *log_fptr = NULL;
static int process_argc;
static char **process_argv;
#ifdef HAVE_THREADS
int running_threads;
pthread_mutex_t cache_sync;
pthread_mutex_t whohandler_sync, whohandler_mass_sync;
static pthread_mutex_t log_sync;
#endif

static void check_firstrun();

void cleanup() {
    stop_modules();
    free_sockets();
    qserver_free();
    free_parser();
    free_UserNode();
    free_ChanNode();
    free_bind();
    free_modcmd();
    free_whoqueue();
    free_mysql();
    free_handleinfohandler();
    free_lang();
}

static int load_mysql_config() {
    char *mysql_host, *mysql_user, *mysql_pass, *mysql_base;
    int mysql_serverport;
    
    mysql_host = get_string_field("MySQL.host");
    if(!mysql_host) {
        perror("invalid neonserv.conf: missing MySQL.host");
        return 0;
    }
    mysql_serverport = get_int_field("MySQL.port");
    if(!mysql_serverport)
        mysql_serverport = 3306;
    mysql_user = get_string_field("MySQL.user");
    if(!mysql_user) {
        perror("invalid neonserv.conf: missing MySQL.user");
        return 0;
    }
    mysql_pass = get_string_field("MySQL.pass");
    if(!mysql_pass) {
        perror("invalid neonserv.conf: missing MySQL.pass");
        return 0;
    }
    mysql_base = get_string_field("MySQL.base");
    if(!mysql_base) {
        perror("invalid neonserv.conf: missing MySQL base");
        return 0;
    }
    init_mysql(mysql_host, mysql_serverport, mysql_user, mysql_pass, mysql_base);
    return 1;
}

static TIMEQ_CALLBACK(clear_cache) {
    timeq_add(CLEAR_CACHE_INTERVAL, 0, clear_cache, NULL);
    clearTempUsers();
    destroyEvents();
    mysql_free();
}

void *thread_main(void *arg) {
    while(running) {
        iohandler_poll();
    }
    return NULL;
}

#ifdef HAVE_THREADS
pthread_t *current_threads = NULL;

int getCurrentThreadID() {
    if(!current_threads) return 0;
    int i;
    unsigned int my_tid = (unsigned int) pthread_self_tid();
    for(i = 0; i < running_threads; i++) {
        #ifdef WIN32
        if((unsigned int) current_threads[i].p == my_tid)
        #else
        if((unsigned int) current_threads[i] == my_tid)
        #endif
            return i+1;
    }
    return 0;
}

#endif

void exit_daemon() {
    running = 0;
    if(daemonized) {
        remove(PID_FILE);
    }
    if(log_fptr) {
        fclose(log_fptr);
        log_fptr = NULL;
    }
}

int main(int argc, char *argv[]) {
    int run_as_daemon = 1;
    int c;
    process_argv = argv;
    process_argc = argc;
    printf("NeonServ v%s\n\n", NEONSERV_VERSION);
    struct option options[] = {
        {"show", 1, 0, 's'},
        {"foreground", 0, 0, 'f'},
        {"help", 0, 0, 'h'},
        {"version", 0, 0, 'v'},
        {0, 0, 0, 0}
    };
    while ((c = getopt_long(argc, argv, "s:fvh", options, NULL)) != -1) {
        switch (c) {
        case 's':
            print_loglevel = atoi(optarg);
            break;
        case 'f':
            run_as_daemon = 0;
            break;
        case 'v':
            printf("Version: %s.%d (%s)\n", NEONSERV_VERSION, patchlevel, (strcmp(revision, "") ? revision : "-"));
            printf("Build: #%s %s (%s lines, " COMPILER ")\n", compilation, creation, codelines);
            exit(0);
            break;
        case 'h':
            printf("Usage: ./neonserv [-s loglevel] [-f] [-h|-v]\n");
            printf(" -s, --show           show log lines matching loglevel in stdout.\n");
            printf(" -f, --foreground     run NeonServ in the foreground.\n");
            printf(" -h, --help           prints this usage message.\n");
            printf(" -v, --version        prints this program's version.\n");
            exit(0);
            break;
        }
    }
    #ifndef WIN32
    if(geteuid() == 0 || getuid() == 0) {
        fprintf(stderr, "NeonServ may not be run with super user privileges.\n");
        exit(0);
    }
    #endif
    #ifdef ENABLE_MEMORY_DEBUG
    initMemoryDebug();
    #endif
    if(!loadConfig(CONF_FILE)) {
        fprintf(stderr, "Unable to load " CONF_FILE "\n");
        exit(0);
    }
    init_bind();
    event_reload(1);
    #if HAVE_THREADS
    THREAD_MUTEX_INIT(log_sync);
    #endif
    if(!load_mysql_config()) {
        fprintf(stderr, "Unable to connect to MySQL\n");
        exit(0);
    }
    check_firstrun();
    char **modulelist = get_all_fieldnames("modules");
    if(!modulelist || !modulelist[0]) {
        fprintf(stderr, "no modules loaded... Please update your configuration!\n");
        exit(0);
    }
    free(modulelist);
    if (run_as_daemon) {
        #ifndef WIN32
        /* Attempt to fork into the background if daemon mode is on. */
        pid_t pid = fork();
        if (pid < 0) {
            fprintf(stderr, "Unable to fork: %s\n", strerror(errno));
        } else if (pid > 0) {
            printf("Forking into the background (pid: %d)...\n", pid);
            putlog(LOGLEVEL_INFO, "Forking into the background (pid: %d)...\n", pid);
            exit(0);
        }
        setsid();
        daemonized = 1;
        atexit(exit_daemon);
        FILE *pidfile = fopen(PID_FILE, "w");
        if (pidfile == NULL) {
            fprintf(stderr, "Unable to create PID file: %s\n", strerror(errno));
            putlog(LOGLEVEL_ERROR, "Unable to create PID file: %s\n", strerror(errno));
        } else {
            fprintf(pidfile, "%i\n", (int)getpid());
            fclose(pidfile);
        }
		FILE *retn;
        fclose(stdin); retn = fopen("/dev/null", "r");
        fclose(stdout); retn = fopen("/dev/null", "w");
        fclose(stderr); retn = fopen("/dev/null", "w");
        #endif
    }
    
main:
    signal(SIGABRT, sighandler);
    signal(SIGFPE, sighandler);
    signal(SIGILL, sighandler);
    signal(SIGINT, sighandler);
    signal(SIGSEGV, sighandler);
    signal(SIGTERM, sighandler);
    
    start_time = time(0);
    
    statistics_enabled = get_int_field("statistics.enable");
    
    #ifdef HAVE_THREADS
    THREAD_MUTEX_INIT(cache_sync);
    THREAD_MUTEX_INIT(whohandler_sync);
    THREAD_MUTEX_INIT(whohandler_mass_sync);
    #endif
    
    init_lang();
    init_parser();
    init_UserNode();
    init_ChanNode();
    init_ModeNode();
    init_bind();
	init_modcmd();
    register_module_commands();
    init_handleinfohandler();
    init_tools();
    init_modulefunctions();
    loadModules();
    init_bots();
    init_DBHelper();
    qserver_init();
    
    load_languages();
    int update_minutes = get_int_field("statistics.frequency");
    if(!update_minutes) update_minutes = 2;
    timeq_add(update_minutes * 60 + 10, 0, main_statistics, NULL);
    
    timeq_add(90, 0, main_checkauths, NULL);
    
    timeq_add(CLEAR_CACHE_INTERVAL, 0, clear_cache, NULL);
    
    int worker_threads = get_int_field("General.worker_threads");
    if(!worker_threads) worker_threads = 1;
    running = 1;
    #ifdef HAVE_THREADS
    int tid_id = 0;
    current_threads = calloc(worker_threads, sizeof(*current_threads));
    for(tid_id = 0; tid_id < worker_threads; tid_id++) {
        running_threads++;
        pthread_create(&current_threads[tid_id], NULL, thread_main, NULL);
    }
    #endif
    thread_main(NULL);
    #ifdef HAVE_THREADS
    for(tid_id = 0; tid_id < worker_threads; tid_id++) {
        pthread_join(current_threads[tid_id], NULL);
    }
    running_threads = 0;
    #endif
    cleanup();
    if(hard_restart) {
        restart_process();
    }
    goto main;
}

void restart_process() {
    /* Append a NULL to the end of argv[]. */
    char **restart_argv = (char **)alloca((process_argc + 1) * sizeof(char *));
    memcpy(restart_argv, process_argv, process_argc * sizeof(char *));
    restart_argv[process_argc] = NULL;
    #ifdef WIN32
    execv(process_argv[0], (const char * const*)restart_argv);
    #else
    execv(process_argv[0], restart_argv);
    #endif
}

int stricmp (const char *s1, const char *s2)
{
   if (s1 == NULL) return s2 == NULL ? 0 : -(*s2);
   if (s2 == NULL) return *s1;
   char c1, c2;
   while ((c1 = tolower (*s1)) == (c2 = tolower (*s2)))
   {
     if (*s1 == '\0') break;
     ++s1; ++s2;
   }
   return c1 - c2;
}

int stricmplen (const char *s1, const char *s2, int len)
{
   if (s1 == NULL) return s2 == NULL ? 0 : -(*s2);
   if (s2 == NULL) return *s1;
   char c1, c2;
   int i = 0;
   while ((c1 = tolower (*s1)) == (c2 = tolower (*s2)))
   {
     i++;
     if (*s1 == '\0') break;
     ++s1; ++s2;
     if(i == len) break;
   }
   return c1 - c2;
}

void restart_bot(int do_hard_restart) {
    hard_restart = do_hard_restart;
    running = 0;
}

void stop_bot() {
    cleanup();
    exit(0);
}

void reload_config() {
    loadConfig(CONF_FILE);
    event_reload(0);
}

static int getCurrentSecondsOfDay() {
    time_t now = time(0);
    struct tm *timeofday = localtime(&now);
    int seconds = 0;
    seconds += timeofday->tm_hour * 3600;
    seconds += timeofday->tm_min * 60;
    seconds += timeofday->tm_sec;
    return seconds;
}

static AUTHLOOKUP_CALLBACK(main_checkauths_callback) {
    //check if registered is still valid
    MYSQL_RES *res;
    MYSQL_ROW row;
    printf_mysql_query("SELECT `user_id`, `user_registered` FROM `users` WHERE `user_user` = '%s'", escape_string(auth));
    res = mysql_use();
    if ((row = mysql_fetch_row(res)) != NULL) {
        int diff = registered - atoi(row[1]);
        if(diff < 0)
            diff *= -1;
        if(!exists || (strcmp(row[1], "0") && diff > 86400)) {
            //User is no longer valid! Delete it...
            deleteUser(atoi(row[0]));
            char *alertchan = get_string_field("General.CheckAuths.alertchan");
            if(alertchan) {
                char reason[MAXLEN];
                if(!exists) {
                    strcpy(reason, "USER_NOT_EXISTS");
                } else {
                    sprintf(reason, "USER_REGISTERED_MISSMATCH: %lu, expected %d (diff: %d)", (unsigned long) registered, atoi(row[1]), diff);
                }
                struct ChanNode *alertchan_chan = getChanByName(alertchan);
                struct ClientSocket *alertclient;
                if(alertchan_chan && (alertclient = getChannelBot(alertchan_chan, 0)) != NULL) {
                    putsock(alertclient, "PRIVMSG %s :Deleted User %s (%s)", alertchan_chan->name, auth, reason);
                }
            }
        } else if(exists && !strcmp(row[1], "0")) {
            printf_mysql_query("UPDATE `users` SET `user_registered` = '%lu', `user_lastcheck` = UNIX_TIMESTAMP() WHERE `user_id` = '%s'", (unsigned long) registered, row[0]);
        } else {
            printf_mysql_query("UPDATE `users` SET `user_lastcheck` = UNIX_TIMESTAMP() WHERE `user_id` = '%s'", row[0]);
        }
    }
}

TIMEQ_CALLBACK(main_checkauths) {
    int next_call = 600;
    if(get_int_field("General.CheckAuths.enabled")) {
        int check_start_time = get_int_field("General.CheckAuths.start_time") * 3600;
        int duration = get_int_field("General.CheckAuths.duration") * 60;
        int now = getCurrentSecondsOfDay();
        if(now < check_start_time && check_start_time+duration >= 86400) {
            check_start_time -= 86400;
        }
        if(now >= check_start_time && now < (check_start_time + duration)) {
            next_call = get_int_field("General.CheckAuths.interval");
            //get the "longest-unchecked-user"
            MYSQL_RES *res;
            MYSQL_ROW row;
            int lastcheck;
            time_t unixtime = time(0);
            int min_unckecked = get_int_field("General.CheckAuths.min_unckecked");
            printf_mysql_query("SELECT `user_user`, `user_lastcheck` FROM `users` ORDER BY `user_lastcheck` ASC LIMIT 1");
            res = mysql_use();
            if ((row = mysql_fetch_row(res)) != NULL) {
                lastcheck = atoi(row[1]);
                if(!lastcheck || unixtime - lastcheck >= min_unckecked) {
                    lookup_authname(row[0], 0, main_checkauths_callback, NULL);
                } else 
                    next_call = 300;
            }
        } else {
            int pending;
            if(now > check_start_time)
                pending = 86400 - now + check_start_time;
            else
                pending = check_start_time - now;
            if(pending < 600)
                next_call = pending;
        }
        
    }
    timeq_add(next_call, 0, main_checkauths, NULL);
}

TIMEQ_CALLBACK(main_statistics) {
    int update_minutes = get_int_field("statistics.frequency");
    if(!update_minutes) update_minutes = 2;
    timeq_add(update_minutes * 60, 0, main_statistics, NULL);
    if(get_int_field("statistics.enable")) {
        statistics_enabled = 1;
        statistics_requested_lusers = 1;
        if(get_int_field("statistics.include_lusers")) {
            struct ClientSocket *bot, *lusersbot = NULL;
            for(bot = getBots(SOCKET_FLAG_READY, NULL); bot; bot = getBots(SOCKET_FLAG_READY, bot)) {
                if(bot->flags & SOCKET_FLAG_PREFERRED)
                    lusersbot = bot;
            }
            bot = lusersbot;
            if(bot == NULL) bot = getBots(SOCKET_FLAG_READY, NULL);
            putsock(bot, "LUSERS");
        } else {
            statistics_update();
        }
    } else
        statistics_enabled = 0;
}

int statistics_update() {
    if(get_int_field("statistics.enable") && statistics_requested_lusers && get_string_field("statistics.execute")) {
        statistics_requested_lusers = 0;
        char command[MAXLEN];
        /* parameters:
         - visible users
         - visible chanusers
         - visible channels
         - privmsg per minute
         - commands per minute
         - network users
         - network channels
        */
        sprintf(command, "%s %d %d %d %d %d %d %d", get_string_field("statistics.execute"), getUserCount(), getChanUserCount(), getChannelCount(), statistics_privmsg, statistics_commands, statistics_network_users, statistics_network_channels);
        statistics_privmsg = 0;
        statistics_commands = 0;
        return system(command);
    }
	return -1;
}

time_t getStartTime() {
    return start_time;
}

int getRunningThreads() {
	#ifdef HAVE_THREADS
    return running_threads;
	#else
	return 1;
	#endif
}

void write_log(int loglevel, const char *line, int len) {
    SYNCHRONIZE(log_sync);
    if(!daemonized && (print_loglevel & loglevel)) {
        printf("%s", line);
    } else if(!daemonized && loglevel == LOGLEVEL_ERROR) {
        fprintf(stderr, "%s", line);
    }
    if(get_int_field("log.loglevel") & loglevel) {
        if(!log_fptr) {
            log_fptr = fopen(LOG_FILE, "a");
            if(!log_fptr) goto write_log_end;
        }
        time_t rawtime;
        struct tm *timeinfo;
        time(&rawtime);
        timeinfo = localtime(&rawtime);
        char timestr[20];
        int timepos = strftime(timestr, 20, "%x %X ", timeinfo);
        fwrite(timestr, 1, timepos, log_fptr);
        fwrite(line, 1, len, log_fptr);
    }
    write_log_end:
    DESYNCHRONIZE(log_sync);
}

void putlog(int loglevel, const char *text, ...) {
    va_list arg_list;
    char logBuf[MAXLOGLEN];
    int pos;
    logBuf[0] = '\0';
    va_start(arg_list, text);
    pos = vsnprintf(logBuf, MAXLOGLEN - 1, text, arg_list);
    va_end(arg_list);
    if (pos < 0 || pos > (MAXLOGLEN - 1)) pos = MAXLOGLEN - 1;
    logBuf[pos] = '\0';
    write_log(loglevel, logBuf, pos);
}

static void check_firstrun() {
    MYSQL_RES *res;
    MYSQL_ROW row;
    printf_mysql_query("SELECT `user_id` FROM `users` WHERE `user_access` = 1000 LIMIT 1");
    res = mysql_use();
    if (mysql_fetch_row(res) == NULL) {
        //first run!
        printf("No superuser found...\n");
        check_firstrun_admin:
        printf("AuthServ account name of admin user: ");
        char *ptr;
        char adminuser[31];
        ptr = fgets(adminuser, 30, stdin);
        for(ptr = adminuser; *ptr; ptr++) { if(*ptr == '\n' || *ptr == '\r') *ptr = '\0'; }
        if(strlen(adminuser) < 2) goto check_firstrun_admin;
        printf_mysql_query("SELECT `user_id` FROM `users` WHERE `user_user` = '%s'", escape_string(adminuser));
        res = mysql_use();
        if ((row = mysql_fetch_row(res)) != NULL)
            printf_mysql_query("UPDATE `users` SET `user_access` = 1000 WHERE `user_id` = '%s'", row[0]);
        else
            printf_mysql_query("INSERT INTO `users` (`user_user`, `user_access`) VALUES ('%s', 1000)", escape_string(adminuser));
    }
    printf_mysql_query("SELECT `id` FROM `bots` WHERE `active` = 1 LIMIT 1");
    res = mysql_use();
    if (mysql_fetch_row(res) == NULL) {
        //no bot active
        printf("No active bot found...\n\n");
        printf("ADD NEW BOT\n");
        char *ptr;
        char bot_nick[31];
        check_firstrun_bot_nick:
        printf("Nick: ");
        ptr = fgets(bot_nick, 30, stdin);
        for(ptr = bot_nick; *ptr; ptr++) { if(*ptr == '\n' || *ptr == '\r') *ptr = '\0'; }
        if(strlen(bot_nick) < 2) goto check_firstrun_bot_nick;
        char bot_ident[16];
        check_firstrun_bot_ident:
        printf("Ident: ");
        ptr = fgets(bot_ident, 15, stdin);
        for(ptr = bot_ident; *ptr; ptr++) { if(*ptr == '\n' || *ptr == '\r') *ptr = '\0'; }
        if(strlen(bot_ident) < 2) goto check_firstrun_bot_ident;
        char bot_realname[101];
        check_firstrun_bot_realname:
        printf("Realname: ");
        ptr = fgets(bot_realname, 100, stdin);
        for(ptr = bot_realname; *ptr; ptr++) { if(*ptr == '\n' || *ptr == '\r') *ptr = '\0'; }
        if(strlen(bot_realname) < 2) goto check_firstrun_bot_realname;
        char bot_server[101];
        check_firstrun_bot_server:
        printf("Server: [irc.onlinegamesnet.net] ");
        ptr = fgets(bot_server, 100, stdin);
        for(ptr = bot_server; *ptr; ptr++) { if(*ptr == '\n' || *ptr == '\r') *ptr = '\0'; }
        if(*bot_server && strlen(bot_nick) < 5) goto check_firstrun_bot_server;
        if(!*bot_server)
            strcpy(bot_server, "irc.onlinegamesnet.net");
        int bot_port;
        char bot_port_buf[7];
        printf("Port: [6667] ");
        ptr = fgets(bot_port_buf, 6, stdin);
        for(ptr = bot_port_buf; *ptr; ptr++) { if(*ptr == '\n' || *ptr == '\r') *ptr = '\0'; }
        if(!*bot_port_buf)
            bot_port = 6667;
        else
            bot_port = atoi(bot_port_buf);
        int bot_ssl;
        char bot_ssl_buf[5];
        check_firstrun_bot_ssl:
        printf("SSL: [y/N] ");
        ptr = fgets(bot_ssl_buf, 4, stdin);
        for(ptr = bot_ssl_buf; *ptr; ptr++) { if(*ptr == '\n' || *ptr == '\r') *ptr = '\0'; }
        if(!*bot_ssl_buf || tolower(*bot_ssl_buf) == 'n')
            bot_ssl = 0;
        else if(tolower(*bot_ssl_buf) == 'y')
            bot_ssl = 1;
        else
            goto check_firstrun_bot_ssl;
        char bot_pass[101];
        printf("Server Password: [] ");
        ptr = fgets(bot_pass, 100, stdin);
        for(ptr = bot_pass; *ptr; ptr++) { if(*ptr == '\n' || *ptr == '\r') *ptr = '\0'; }
        int bot_maxchan;
        char bot_maxchan_buf[5];
        printf("MaxChannel: [20] ");
        ptr = fgets(bot_maxchan_buf, 5, stdin);
        for(ptr = bot_maxchan_buf; *ptr; ptr++) { if(*ptr == '\n' || *ptr == '\r') *ptr = '\0'; }
        if(*bot_maxchan_buf)
            bot_maxchan = atoi(bot_maxchan_buf);
        else
            bot_maxchan = 20;
        int bot_queue;
        char bot_queue_buf[5];
        check_firstrun_bot_queue:
        printf("Queue (prevents excess floods): [Y/n] ");
        ptr = fgets(bot_queue_buf, 4, stdin);
        for(ptr = bot_queue_buf; *ptr; ptr++) { if(*ptr == '\n' || *ptr == '\r') *ptr = '\0'; }
        if(!*bot_queue_buf || tolower(*bot_queue_buf) == 'y')
            bot_queue = 1;
        else if(tolower(*bot_queue_buf) == 'n')
            bot_queue = 0;
        else
            goto check_firstrun_bot_queue;
        printf_mysql_query("INSERT INTO `bots` (`active`, `nick`, `server`, `port`, `pass`, `ssl`, `ident`, `realname`, `botclass`, `textbot`, `queue`, `defaulttrigger`, `max_channels`) VALUES ('1', '%s', '%s', '%d', '%s', '%d', '%s', '%s', '1', '1', '%d', '+', '%d')", escape_string(bot_nick), escape_string(bot_server), bot_port, escape_string(bot_pass), bot_ssl, escape_string(bot_ident), escape_string(bot_realname), bot_queue, bot_maxchan);
    }
}

